--------------------------------------------------------------------------------
-- Title      : operativeUnit
-- Project    :
-------------------------------------------------------------------------------
-- File       : operativeUnit.vhd
-- Author     : Jean-Noel BAZIN  <jnbazin@pc-disi-026.enst-bretagne.fr>
-- Company    :
-- Created    : 2018-04-11
-- Last update: 2019-02-13
-- Platform   :
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Operative unit of a sequential FIR filter. Including shift
-- register for samples, registers for coefficients, a MAC and a register to
-- store the result
-------------------------------------------------------------------------------
-- Copyright (c) 2018
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2019-02-13  1.1      marzel  Update to provide a 16-tap filter and improve
--                              the user experience ;)
-- 2018-04-11  1.0      jnbazin Created
-- 2018-04-18  1.0      marzel  Modification of SR_Y assignment to a round
--                              instead of a trunc
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity OperativeUnit is

  port (
    I_clock          : in  std_logic;   -- global clock
    I_reset          : in  std_logic;   -- asynchronous global reset
    I_inputSample    : in  std_logic_vector(10 downto 0);  -- 
    I_loadShift1     : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_loadShift2     : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_loadShift3     : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_initAddress    : in  std_logic;  -- Control signal to initialize register read address
    I_incrAddress    : in  std_logic;  -- Control signal to increment register read address
    I_initSum        : in  std_logic;  -- Control signal to initialize the MAC register
    I_loadSum        : in  std_logic;  -- Control signal to load the MAC register;
    I_loadY          : in  std_logic;   -- Control signal to load Y register
    I_Filter         : in std_logic_vector(1 downto 0);
    I_FilteredSampleDone : in std_logic;
    O_processingDone1 : out std_logic;   -- Indicate that processing is done
    O_processingDone2 : out std_logic;   -- Indicate that processing is done
    O_processingDone2b : out std_logic;   -- Indicate that processing is done
    O_processingDone3 : out std_logic;   -- Indicate that processing is done
    O_Y              : out std_logic_vector(10 downto 0)   -- filtered sample
    );

end entity OperativeUnit;

architecture arch_OperativeUnit of OperativeUnit is
  type registerFile95 is array(0 to 94) of signed(10 downto 0);
  signal SR_coefRegister1 : registerFile95;
  type registerFile3 is array(0 to 2) of signed(10 downto 0);
  signal SR_coefRegister2 : registerFile3;
  type registerFile2 is array(0 to 1) of signed(10 downto 0);
  signal SR_coefRegister2b : registerFile2;
  type registerFile11 is array(0 to 10) of signed(10 downto 0);
  signal SR_coefRegister3 : registerFile11;


  signal SR_shiftRegister1 : registerFile95;  -- shift register file used to store and shift input samples
  signal SR_shiftRegister2 : registerFile3;  -- shift register file used to store and shift input samples
  signal SR_shiftRegister3 : registerFile11;  -- shift register file used to store and shift input samples
  signal SC_multOperand1  : signed(10 downto 0);
  signal SC_multOperand2  : signed(10 downto 0);
  signal SC_MultResult    : signed(21 downto 0);  -- Result of the multiplication Xi*Hi
  signal SC_addResult     : signed(28 downto 0);  -- result of the accumulation addition
  signal SR_sum           : signed(28 downto 0);  -- Accumulation register
  signal SR_Y             : signed(10 downto 0);  -- filtered sample storage register
  signal SR_readAddress   : integer range 0 to 94;  -- register files read address

-- DIMENSIONNEMENT !!!

begin 

-- Low-pass filter provided with octave (or Matlab ;)) command
--fir1(15, .001)/sqrt(sum(fir1(15, .001).^2))*2^6
  SR_coefRegister1 <= (to_signed(-1, 11),  
                        to_signed(-1, 11),
                        to_signed(-1, 11),
                        to_signed(-1, 11),
                        to_signed(-1, 11),
                        to_signed(-1, 11),
                        to_signed(-2, 11),
                        to_signed(-2, 11),
                        to_signed(-2, 11),
                        to_signed(-3, 11),
                        to_signed(-3, 11),
                        to_signed(-3, 11),
                        to_signed(-4, 11),
                        to_signed(-4, 11),
                        to_signed(-5, 11),
                        to_signed(-5, 11),
                        to_signed(-6, 11),
                        to_signed(-6, 11),
                        to_signed(-7, 11),
                        to_signed(-7, 11),
                        to_signed(-8, 11),
                        to_signed(-8, 11),
                        to_signed(-9, 11),
                        to_signed(-10, 11),
                        to_signed(-10, 11),
                        to_signed(-11, 11),
                        to_signed(-11, 11),
                        to_signed(-12, 11),
                        to_signed(-13, 11),
                        to_signed(-13, 11),
                        to_signed(-14, 11),
                        to_signed(-14, 11),
                        to_signed(-15, 11),
                        to_signed(-15, 11),
                        to_signed(-16, 11),
                        to_signed(-16, 11),
                        to_signed(-17, 11),
                        to_signed(-17, 11),
                        to_signed(-18, 11),
                        to_signed(-18, 11),
                        to_signed(-18, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(1004, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(-19, 11),
                        to_signed(-18, 11),
                        to_signed(-18, 11),
                        to_signed(-18, 11),
                        to_signed(-17, 11),
                        to_signed(-17, 11),
                        to_signed(-16, 11),
                        to_signed(-16, 11),
                        to_signed(-15, 11),
                        to_signed(-15, 11),
                        to_signed(-14, 11),
                        to_signed(-14, 11),
                        to_signed(-13, 11),
                        to_signed(-13, 11),
                        to_signed(-12, 11),
                        to_signed(-11, 11),
                        to_signed(-11, 11),
                        to_signed(-10, 11),
                        to_signed(-10, 11),
                        to_signed(-9, 11),
                        to_signed(-8, 11),
                        to_signed(-8, 11),
                        to_signed(-7, 11),
                        to_signed(-7, 11),
                        to_signed(-6, 11),
                        to_signed(-6, 11),
                        to_signed(-5, 11),
                        to_signed(-5, 11),
                        to_signed(-4, 11),
                        to_signed(-4, 11),
                        to_signed(-3, 11),
                        to_signed(-3, 11),
                        to_signed(-3, 11),
                        to_signed(-2, 11),
                        to_signed(-2, 11),
                        to_signed(-2, 11),
                        to_signed(-1, 11),
                        to_signed(-1, 11),
                        to_signed(-1, 11),
                        to_signed(-1, 11),
                        to_signed(-1, 11),
                        to_signed(-1, 11)
                      );
                      
 SR_coefRegister2 <= (to_signed(961, 11),  -- ROM register used file to store FIR coefficients
                      to_signed(-1554, 11),
                      to_signed(961, 11)
                      );

 SR_coefRegister2b <= (to_signed(-1554, 11),  -- ROM register used file to store FIR coefficients
                      to_signed(897, 11)
                      );
                      
 SR_coefRegister3 <= (to_signed(-119, 11),  -- ROM register used file to store FIR coefficients
                      to_signed(122, 11),
                      to_signed(149, 11),
                      to_signed(191, 11),
                      to_signed(226, 11),
                      to_signed(239, 11),
                      to_signed(226,11),
                      to_signed(191, 11),
                      to_signed(149,11),
                      to_signed(122,11),
                      to_signed(-119,11)
                      );
                   
                      
  shift : process (I_clock, I_reset) is
  begin  -- process shift
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_shiftRegister1 <= (others => (others => '0'));
      SR_shiftRegister2 <= (others => (others => '0'));
      SR_shiftRegister3 <= (others => (others => '0'));
    elsif (rising_edge(I_clock)) then
        if (I_loadShift1 = '1') then
            SR_shiftRegister1(1 to 94) <= SR_shiftRegister1(0 to 93);
            SR_shiftRegister1(0) <= signed(I_inputSample);
        elsif (I_loadShift2 = '1') then
            SR_shiftRegister2(1 to 2) <= SR_shiftRegister2(0 to 1);
            SR_shiftRegister2(0) <= signed(SR_Y);
        elsif (I_loadShift3 = '1') then
            SR_shiftRegister3(1 to 10) <= SR_shiftRegister3(0 to 9);
            SR_shiftRegister3(0) <= signed(SR_Y);
        end if; 
    end if;
  end process shift;

  incr_address : process (I_reset, I_clock) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_readAddress <= 0;
    elsif (rising_edge(I_clock)) then
        if(I_initAddress = '1') then
            SR_readAddress <= 0;
        end if;
            
        if (I_incrAddress = '1') then
            if (I_Filter = "00") then        
                if (SR_readAddress < 94) then
                    SR_readAddress <= SR_readAddress + 1;
                else
                    SR_readAddress <= 0;
                end if;
            elsif (I_Filter = "01") then        
                if (SR_readAddress < 2) then
                    SR_readAddress <= SR_readAddress + 1;
                else
                    SR_readAddress <= 0;
                end if;
            elsif (I_Filter = "10") then        
                if (SR_readAddress < 1) then
                    SR_readAddress <= SR_readAddress + 1;
                else
                    SR_readAddress <= 0;
                end if;
            else        
                if (SR_readAddress < 10) then
                    SR_readAddress <= SR_readAddress + 1;
                else
                    SR_readAddress <= 0;
                end if;
           end if;

        end if;
     end if;
  end process incr_address;
 
  O_processingDone1  <= '1' when (SR_readAddress >=94 and I_Filter="00") else '0' ;
  O_processingDone2  <= '1' when (SR_readAddress >=2 and I_Filter="01") else '0' ;
  O_processingDone2b <= '1' when (SR_readAddress >=1 and I_Filter="10") else '0' ;
  O_processingDone3  <= '1' when (SR_readAddress >=10 and I_Filter="11") else '0' ;

  SC_multOperand1 <= SR_shiftRegister1(SR_readAddress) when I_Filter="00" else SR_shiftRegister2(SR_readAddress) when I_Filter="01" else SR_shiftRegister3(SR_readAddress); 
  SC_multOperand2 <= SR_coefRegister1(SR_readAddress) when I_Filter="00" else SR_coefRegister2(SR_readAddress) when I_Filter="01" else SR_coefRegister2b(SR_readAddress) when I_Filter="10" else SR_coefRegister3(SR_readAddress) ;    -- 8 bits
  SC_MultResult   <= SC_multOperand1 * SC_multOperand2  ;
  SC_addResult    <= resize(SC_MultResult, SC_addResult'length) + SR_sum;

  sum_acc : process (I_reset, I_clock) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_sum <= (others => '0');
    elsif rising_edge(I_clock) then
        if (I_initSum='1') then
            SR_sum <= (others => '0');
        elsif (I_loadSum='1') then
            SR_sum <= SC_addResult;
        end if;
    end if;
  end process sum_acc;

  store_result : process (I_clock) is
  begin
      if (rising_edge(I_clock)) then
        if (I_loadY='1') then
            if(SR_sum(10)='1') then
                SR_Y <= SR_sum(21 downto 11) + 1;
            else
                SR_Y <= SR_sum(21 downto 11);
           end if;
        end if;
        
        if (I_FilteredSampleDone = '1') then
            O_Y <= std_logic_vector(SR_Y);
        end if;
    end if;

  end process store_result;

  

end architecture arch_OperativeUnit;